// -*- C++ -*-
// $Header: /u/drspeech/repos/quicknet2/QN_MLPWeightFile_MLPW.h,v 1.3 2006/01/05 20:37:50 davidj Exp $
// 
// QN_MLPWeightFle_MLPW.h
//
// Define a new *binary* MLP Weight file format
// 1997jun19 dpwe@icsi.berkeley.edu
// Modified significantly, davidj@icsi.berkeley.edu, 2005/2006
//


#ifndef QN_MLPWeightFile_MLPW_h_INCLUDED
#define QN_MLPWeightFile_MLPW_h_INCLUDED


/* Must include the config.h file first */
#include <QN_config.h>
#include <stdio.h>
#include "QN_types.h"
#include "QN_MLPWeightFile.h"

enum QN_MLPW_dtype {
    QN_MLPW_dt_none = 0,
    QN_MLPW_dt_float = 4, 
    QN_MLPW_dt_double = 8
};

// Net types
enum QN_MLPW_ntype {
    QN_MLPW_nt_none = 0,
    QN_MLPW_nt_sigmoid_softmax = 1, // sigmoid on hidden, softmax on output
    QN_MLPW_nt_sigmoid_sigmoid = 2, // sigmoid on hidden, sigmoid on output
    QN_MLPW_nt_sigmoid_linear = 3	// sigmoid on hidden, linear on output
};

// Convert a dtype constant into bytes-per-value
#define QN_MLPW_dt_sizeof(type) (type & 0x0F)

// Magic number at start of file
#define QN_MLPW_MAGIC "MLPW"
#define QN_MLPW_SIZEOFMAGIC 4

#define QN_MLPW_VERSION (20051111)

// Limit so we can use a fixed array for nlayers
#define QN_MLPW_MAX_NLAYERS	5
#define QN_MLPW_MAX_NSECTS	2*(QN_MLPW_MAX_NLAYERS-1)

/* Header structure for MLPW files */
typedef struct _MLPW_hdr {
    QNUInt32 	magic;		// "MLPW" magic number
    QNUInt32 	version;	// version field
    QNUInt32	nettype;	// flag indicating nettype (softmax etc)
    QNUInt32	datatype;	// QN_MLPW_dt_* field
    QNUInt32	nlayers;	// total number of layers (2 -> 5)
    QNUInt32	nsections;	// total number of sections (MLP3 = 4)
    // On disk, this header is followed by:
    // QNUInt32   layersize[nlayers]      number of units in each layer
    // then the (4) sections...
} QN_MLPW_hdr_str;

typedef struct _MLPW_sect_hdr {
    // This is the start of each section
    QNUInt32	sectiontype;	// QN_SectionSelector value
    QNUInt32	numvalues;	// how many values
    // then the coded values
} QN_MLPW_sect_hdr;

class QN_MLPWeightFile_MLPW : public QN_MLPWeightFile
{
  public:
    // If n_layers is nonzero for read, it is checked against the file
    QN_MLPWeightFile_MLPW(int dbg, FILE* stream, QN_FileMode mode,
			  const char* name,
			  size_t n_layers, size_t layer_sizes[], 
			  QN_MLPW_ntype nt = QN_MLPW_nt_none,
			  QN_MLPW_dtype dt = QN_MLPW_dt_double);

    // How many layers are there?
    size_t num_layers() { return nLayers; }

    // How many sections?
    size_t num_sections() { return nSections; }

    // How many units are there in the given layer?
    size_t size_layer(QN_LayerSelector layer) { assert(layer < nLayers); 
                                                return layerSize[layer]; }

    // Return whether weight matrices are input or output major
    enum QN_WeightMaj get_weightmaj(QN_SectionSelector);

    // Return the order of the weights in the file
    enum QN_SectionSelector get_weighttype(int section);

    // Is this fine QN_READ or QN_WRITE?
    enum QN_FileMode get_filemode();

    // Read in a buffer full of weights
    size_t read(float* buf, size_t count);

    // Write out a buffer full of weights
    size_t write(const float* buf, size_t count);
    
private:
    int netType;	// softmax etc
    int nLayers;
    int layerSize[QN_MLPW_MAX_NLAYERS];
    int nSections;

    QN_MLPW_dtype dtype;		// prevailing data type
    int exp;		// prevailing exp, passed to vx conv routines

    int io_count;
    int io_state;

private:
    int debug;			// Level of verbosity
    const enum QN_FileMode mode;
    FILE* const stream;		// Stream used to access weights files
    const char* filename;	// Name of the weight file

    static const char* CLASSNAME; // Name of this class


    int read_header(void);
    int write_header(void);
    int read_sect_header(int sect, int numvals);
    int write_sect_header(int sect, int numvals);
};

#endif

